import * as d3 from 'd3';

import { graph } from './graph';
import { body } from './body';
import { bodyData, organDict } from '@ssmadha/body';
import {
  asif_melted,
  gene_cluster,
  hi_union,
  transcript_data,
} from '@ssmadha/asif';
import {
  parseData,
  parsePackData,
  parseEdges,
} from './parse';
import { simulate } from './simulate';
import { clusterLegend } from './clusterLegend';
import { nodeLegend } from './nodeLegend';
import { infoView } from './infoView';
import './styles.css';

export const main = (container, { state, setState }) => {
  const width = container.clientWidth;
  const height = container.clientHeight;

  const {
    transform,
    nodes,
    radiusScale,
    colorFillScale,
    colorStrokeScale,
    packData,
    links,
    keepData,
    freezeSim,
    clickedCluster,
    hoveredCluster,
    clickedOrgan,
    hoveredOrgan,
    clickedNodes,
    filterLimit,
  } = state;

  if (filterLimit === undefined) {
    setState((state) => ({
      ...state,
      filterLimit: 20,
    }));
    return;
  }

  const nodeLegendWidth = 100;
  const nodeLegendHeight = 500;

  const bodyWidth = 360;
  const bodyHeight = 550;

  const netGraphX = bodyWidth;
  const netGraphY = 0;
  const netGraphWidth = width - netGraphX - nodeLegendWidth;
  const netGraphHeight = bodyHeight;

  const infoHeight = 400;
  const simMinX = 10;
  const simMaxX = netGraphWidth - 10;
  const simMinY = 10;
  const simMaxY = netGraphHeight - 10;

  //zooming in and out once gets the transform scale to 0.999...
  const zoomNameLimit = 1.5;

  //based on https://vizhub.com/curran/multiple-menus-with-d3 and https://vizhub.com/curran/menu-with-d3
  const containerSelection = d3
    .select(container)
    .attr('class', 'ui-elements-container');

  const menusContainer = containerSelection
    .selectAll('div.menus-container')
    .data([null])
    .join('div')
    .attr('class', 'menus-container');

  const filterRangeDiv = menusContainer
    .selectAll('div.filter-range-container')
    .data([null])
    .join('div')
    .attr('class', 'filter-range-container');

  filterRangeDiv
    .selectAll('label')
    .data([null])
    .join('label')
    .attr('class', 'form-label')
    .attr('for', 'filterRange')
    .text(`Expression Limit: ${filterLimit}`);

  filterRangeDiv
    .selectAll('input')
    .data([null])
    .join('input')
    .attr('type', 'range')
    .attr('class', 'form-range')
    .attr('id', 'filterRange')
    .attr(
      'min',
      d3.min(asif_melted, (d) => d.Expression),
    )
    .attr(
      'max',
      d3.max(asif_melted, (d) => d.Expression),
    )
    .attr('value', filterLimit)
    .on('input', (event) =>
      setState((state) => ({
        ...state,
        filterLimit: event.target.value,
        freezeSim: true,
      })),
    )
    .on('change', (event) =>
      setState((state) => ({
        ...state,
        filterLimit: event.target.value,
        keepData: false,
        freezeSim: false,
      })),
    );

  const setClickedCluster = (clickedCluster) => {
    setState((state) => ({
      ...state,
      clickedCluster,
      keepData: true,
      freezeSim: true,
    }));
  };

  const setHoveredCluster = (hoveredCluster) => {
    setState((state) => ({
      ...state,
      hoveredCluster,
      keepData: true,
      freezeSim: true,
    }));
  };

  const clusterLegendDiv = menusContainer
    .selectAll('div.cluster-legend')
    .data([null])
    .join('div')
    .attr('class', 'cluster-legend');

  if (nodes) {
    const clusterLegendSvg = clusterLegendDiv
      .selectAll('svg.cluster-legend')
      .data([null])
      .join('svg')
      .attr('class', 'cluster-legend')
      .attr('width', 360)
      .attr('height', 50);

    clusterLegend(clusterLegendSvg, {
      colorScale: colorStrokeScale,
      legendLabel: 'Cluster',
      legendLabelY: 20,
      legendPadding: 70,
      tickSpacing: 25,
      legendCircleRadius: 5,
      tickPadding: 10,
      hoveredCluster,
      setHoveredCluster,
      clickedCluster,
      setClickedCluster,
    });
  }

  const setClickedOrgan = (clickedOrgan) => {
    setState((state) => ({
      ...state,
      transform: undefined,
      clickedOrgan,
      keepData: false,
      freezeSim: false,
    }));
  };

  const setHoveredOrgan = (hoveredOrgan) => {
    setState((state) => ({
      ...state,
      hoveredOrgan,
      keepData: true,
      freezeSim: true,
    }));
  };

  const viewDiv = containerSelection
    .selectAll('div.view-container')
    .data([null])
    .join('div')
    .attr('class', 'flex-row view-container');

  const bodyGroup = viewDiv
    .selectAll('div.body')
    .data([null])
    .join('div')
    .attr('class', 'body');

  const bodySVG = bodyGroup
    .selectAll('svg.body')
    .data([null])
    .join('svg')
    .attr('class', 'body')
    .attr('width', bodyWidth)
    .attr('height', bodyHeight)
    //.attr('preserveAspectRatio', 'none')
    .style('background', null);

  bodySVG.call(body, {
    bodyData,
    organDict,
    legendLabel: 'Organ',
    legendLabelFontSize: 22,
    legendX: 20,
    legendY: 25,
    legendPadding: 13,
    legendFontSize: 12,
    tickSpacing: 13,
    tickPadding: 0,
    clickedOrgan,
    setClickedOrgan,
    hoveredOrgan,
    setHoveredOrgan,
  });

  const tissue = clickedOrgan;

  if (!keepData) {
    if (clickedOrgan) {
      const filteredData = asif_melted.filter(
        (d) =>
          d.Tissue === tissue &&
          d.Expression >= filterLimit,
      );
      const {
        nodes,
        radiusScale,
        colorFillScale,
        colorStrokeScale,
      } = parseData({
        data: filteredData,
        id: (d) => d['Gene ID'],
        text: (d) => d['Gene Name'],
        simMinX,
        simMaxX,
        simMinY,
        simMaxY,
        radiusValue: (d) => d.Expression,
        radiusExtent: [5, 15],
        nodeFillValue: (d) => d.ASIF,
        nodeStrokeValue: gene_cluster[0],
      });
      setState((state) => ({
        ...state,
        nodes,
        radiusScale,
        colorFillScale,
        colorStrokeScale,
        links: parseEdges({
          data: filteredData,
          dataID: (d) => d['Gene ID'],
          edges: hi_union,
          sourceValue: (d) => d.source,
          targetValue: (d) => d.target,
          edgeColorValue: (d) => d.coverage,
        }),
        packData: parsePackData({
          data: filteredData,
          dataID: (d) => d['Gene ID'],
          subData: transcript_data.filter(
            (d) => d.Tissue == tissue,
          ),
          parentID: (d) => d.Gene,
          childID: (d) => d.Transcript,
          radiusValue: (d) => d.Expression,
          colorValue: (d) => d.ASIF,
        }),
        keepData: true,
      }));
      return;
    } else {
      setState((state) => ({
        ...state,
        nodes: null,
        links: null,
        keepData: true,
      }));
    }
  }

  const graphDiv = viewDiv
    .selectAll('div.graph-container')
    .data([null])
    .join('div')
    .attr('class', 'graph-container');

  const graphSvg = graphDiv
    .selectAll('svg.graph')
    .data([null])
    .join('svg')
    .attr('class', 'graph')
    .attr('x', netGraphX)
    .attr('y', netGraphY)
    .attr('width', netGraphWidth)
    .attr('height', netGraphHeight);

  if (!freezeSim && clickedOrgan) {
    simulate({
      nodes,
      links,
      netGraphWidth,
      netGraphHeight,
      simMinX,
      simMaxX,
      simMinY,
      simMaxY,
      linkDistance: 60,
      linkIterations: 10,
      manyBodyStrength: -150,
      manyBodyMinDist: 1,
      manyBodyMaxDist: (d) => d.r + 60,
      collideRadius: (d) => d.r + 15,
      centerX: netGraphWidth / 2,
      centerY: netGraphHeight / 2,
      centerStrength: 0.6,
      strengthXY: 0.5,
      velocityDecay: 0.8,
    });
  }

  //graphSvg.attr('transform', transform);

  //zoom code from https://vizhub.com/curran/d6f1170765c84a498caa6ea11403e3be
  const zoomBehavior = d3
    .zoom()
    .scaleExtent([0.75, 15])
    .on('zoom', (event) => {
      setState((state) => ({
        ...state,
        transform: event.transform,
        freezeSim: true,
      }));
    });

  const setClickedNode = (clickedNode) => {
    setState((state) => ({
      ...state,
      clickedNodes: [clickedNode],
      keepData: true,
      freezeSim: true,
    }));
    return;
    if (!clickedNodes || clickedNodes.length == 0) {
      console.log(clickedNodes);
      console.log('empty clicked nodes');
      setState((state) => ({
        ...state,
        clickedNodes: [clickedNode],
        keepData: true,
        freezeSim: true,
      }));
    } else if (clickedNodes.includes(clickedNode)) {
      console.log('removing');
      setState((state) => ({
        ...state,
        clickedNodes: clickedNodes.filter(
          (d) => d.id !== clickedNode.id,
        ),
        keepData: true,
        freezeSim: true,
      }));
    } else {
      console.log('pushing');
      clickedNodes.push(clickedNode);
      setState((state) => ({
        ...state,
        clickedNodes: [clickedNode],
        keepData: true,
        freezeSim: true,
      }));
    }
  };

  graphSvg.call(zoomBehavior);
  //console.log(nodes);
  graph(graphSvg, {
    nodes,
    links,
    packData,
    nodeStrokeWidth: 1,
    linkStrokeWidth: 1,
    textFontSize: 16,
    transform,
    zoomNameLimit,
    hoveredCluster,
    clickedCluster,
    setClickedNode,
    setState,
  });

  const nodeLegendDiv = viewDiv
    .selectAll('div.nodeLegend')
    .data([null])
    .join('div')
    .attr('class', 'nodeLegend');

  const nodeLegendSvg = nodeLegendDiv
    .selectAll('svg.nodeLegend')
    .data([null])
    .join('svg')
    .attr('class', 'nodeLegend')
    .attr('width', nodeLegendWidth)
    .attr('height', nodeLegendHeight);

  if (nodes) {
    nodeLegend(nodeLegendSvg, {
      colorScale: colorFillScale,
      colorLabel: 'ASIF score',
      sizeScale: radiusScale,
      sizeLabel: 'Expression',
      legendLabelFontSize: 16,
      legendPadding: 20,
      legendX: 10,
      legendY: 15,
      legendFontSize: 12,
      colorNumber: 100,
      colorPadding: 15,
      colorHeight: 250,
      colorWidth: 20,
      colorTickNumber: 8,
      legendSpacing: 35,
      sizeNumber: 4,
      sizeSpacing: 30,
      sizePadding: 17,
    });
  }

  const infoDiv = containerSelection
    .selectAll('div.info-container')
    .data([null])
    .join('div')
    .attr('class', 'info-container');

  infoView(infoDiv, {
    nodes: clickedNodes,
    data: asif_melted,
    subData: transcript_data,
    tissue,
  });
};
